from btconfig.logger import Logger
from btconfig.configutils import AttrDict
from btconfig.configloader import ConfigLoader
import os
import sys
from jinja2 import Template
import yaml
from string import Template

# Setup Logging
logger = Logger().init_logger(__name__)

class SuperDuperConfig(ConfigLoader):

  def render(self, **kwargs):

    config_content = kwargs.get('config_content')
    config_file_uri = kwargs.get('uri')
    config_dict = {}
    config_is_valid = False
    invalid_keys = []

    try:
      if not config_content:
        _ymlfile_content = open(config_file_uri).read()
      else:
        _ymlfile_content = config_content
      ymlfile_content = Template(_ymlfile_content).safe_substitute(os.environ)
      if self.templatized:
        try:
          ymlfile_template = Template(ymlfile_content)
          ymlfile_data = ymlfile_template.render(
            session=self.initial_data
          )
        except Exception as e:
          logger.warning(
            f"I had trouble rendering the config, \
            error was {e}"
          )
          if self.failfast:
            sys.exit(1)
          else:
            ymlfile_data = ymlfile_content
      else:
        ymlfile_data = ymlfile_content
      cfg = yaml.safe_load(ymlfile_data)
      config_dict = cfg[self.data_key] if self.data_key is not None else cfg
      if isinstance(config_dict, dict):
        config_dict['config_file_uri'] = config_file_uri
      else:
        _config_dict = [e for e in config_dict if isinstance(e, dict)]
        if len(_config_dict) >= 0:
          config_dict = config_dict[0]
          config_dict['config_file_uri'] = config_file_uri
        else:
          logger.warning("Got unexpected data structure from config dictionary, not setting 'config_file_uri' key")
      invalid_keys = [m[m.keys()[0]].get(k) for k in self.req_keys for m in config_dict if m[m.keys()[0]].get(k)]
      config_is_valid = len(invalid_keys) == 0
      self.logger.debug(f"Found config file - {config_file_uri}")
      if not config_is_valid:
        invalid_keys_string = ','.join(invalid_keys)
        self.logger.warning(
          f"The following required keys were not defined \
          in your input file {config_file_uri}: \
          {invalid_keys_string}"
        )
        self.logger.warning(
          "Review the available documentation or consult --help")
    except Exception as e:
      self.logger.warning(
      f"I encountered a problem reading your \
      input file: {config_file_uri}, error was {e}"
      )
    return config_dict, config_is_valid, invalid_keys

  def read(self, **kwargs):
    """Load specified config file"""

    pre_existing_config_data = kwargs.get('pre_existing_config_data')
    config_file_uri = kwargs.get('config_file_uri', self.config_file_uri)
    config_file_auth_username = kwargs.get('config_file_auth_username')
    config_file_auth_password = kwargs.get('config_file_auth_password')

    if config_file_uri in self.configs_already_processed:
      logger.debug(f'Already processed {config_file_uri}, skipping ...')
      return

    if config_file_uri.startswith('http'):
      config_data = self.load_from_web(
      config_file_uri,
      config_file_auth_username=config_file_auth_username,
      config_file_auth_password=config_file_auth_password
      )
    else:
      config_data = self.load(config_file_uri)

    if pre_existing_config_data and not config_data:
      return pre_existing_config_data

    external_configs = config_data.get('external_configs', [])

    if len(external_configs) > 0:
      logger.debug(f'External configs are being referenced in {config_file_uri}')
      for external_config in external_configs:
        if isinstance(external_config, dict):
          config_uri = external_config.get('uri')
          logger.debug(f'Loading {config_uri} ...')
          if config_uri:
            config_uri_username = external_config.get('auth',{}).get('username')
            config_uri_password = external_config.get('auth',{}).get('password')
            templatized = external_config.get('templatized')
            external_settings = self.read(
              config_file_uri=config_uri,
              config_file_auth_username=config_uri_username,
              config_file_auth_password=config_uri_password,
              templatized=templatized,
              initial_data=self.initial_data,
              pre_existing_config_data=config_data
            )
            if isinstance(external_settings, dict):
              config_data = AttrDict.merge(config_data, external_settings)
              return config_data
    else:
      config_data = AttrDict(config_data)
      return config_data


from btconfig.configutils import AttrDict
from btconfig.logger import Logger
import os
import sys

# Setup Logging
logger = Logger().init_logger(__name__)

class ConfigLoader:

    def load_from_web(self, config_file_uri, **kwargs):
        """Load specified config file from http URI"""

        config_file_auth_username = kwargs.get('config_file_auth_username', self.config_file_auth_username)
        config_file_auth_password = kwargs.get('config_file_auth_password', self.config_file_auth_password)

        if config_file_auth_username and config_file_auth_password:
            config_res = self.webadapter.get(
                config_file_uri,
                username=config_file_auth_username,
                password=config_file_auth_password
            )
        else:
            config_res = self.webadapter.get(config_file_uri)

        if not config_res:
            return {}

        config_dict, config_is_valid, invalid_keys = self.render(
            uri=config_file_uri,
            config_content=config_res
        )

        self.configs_already_processed.append(config_file_uri)

        if config_dict:
            return AttrDict(config_dict)
        else:
            return {}

    def load(self, config_file_uri):
        """Load specified config file from filesystem"""

        config_found = False

        if not os.path.exists(config_file_uri):
            config_search_paths = [
                os.path.realpath(os.path.expanduser('~')),
                '.',
                os.path.join(os.path.abspath(os.sep), 'etc')
            ]
            if self.extra_config_search_paths:
                if isinstance(self.extra_config_search_paths, list):
                    config_search_paths += self.extra_config_search_paths
                elif isinstance(self.extra_config_search_paths, str):
                    config_search_paths += [self.extra_config_search_paths]
                else:
                    logger.error(
                        'extra_config_search_paths must \
                        be of type str or list'
                    )
                    sys.exit(1)

            config_file_uris = [
                os.path.expanduser(os.path.join(p, config_file_uri))
                for p in config_search_paths
            ]
        else:
            config_file_uris = [config_file_uri]
            config_found = True
            logger.debug(f'Found config at {config_file_uri}')

        for cf_uri in config_file_uris:
            config_exists = config_found or os.path.exists(cf_uri)
            if config_exists:
                config_found = True
                self.configs_already_processed.append(cf_uri)
                config_dict, config_is_valid, invalid_keys = self.render(
                    uri=cf_uri,
                    templatized=self.templatized,
                    failfast=self.failfast,
                    data_key=self.data_key,
                    initial_data=self.initial_data,
                    req_keys=self.req_keys
                )
                break

        if not config_found and self.failfast:
            self.logger.error(
                "Could not find any config file in \
                specified search paths. Aborting."
            )
            sys.exit(1)

        if config_found and config_is_valid:
            config_data = config_dict
        else:
            if self.failfast:
                self.logger.error(
                    "Aborting due to invalid or not found \
                    config(s) [%s]" % ','.join(config_file_uris)
                )
                sys.exit(1)
            else:
                logger.debug('No settings could be derived, using defaults')
                config_data = self.default_value

        return config_data


from btconfig.logger import Logger
import os
from btconfig.superconf import SuperDuperConfig
from btweb import WebAdapter

# Setup Logging
logger = Logger(is_main_logger=True).init_logger(__name__)

class Config(SuperDuperConfig):

  def __init__(self, **kwargs):

    self.logger = logger
    self.initial_data = kwargs.get('initial_data', {'environment': os.environ})
    self.verify_tls = kwargs.get('verify_tls', False)
    self.webadapter = WebAdapter(verify_tls=self.verify_tls)
    self.args = kwargs.get('args', {})
    self.config_is_templatized = self.args.get('is_template', False) or kwargs.get('is_template', False)
    self.extra_config_search_paths = kwargs.get('extra_config_search_paths', '')
    config_file_uri = kwargs.get('config_file_uri')
    if config_file_uri:
      self.config_file_uri = os.path.expanduser(config_file_uri)
    self.req_keys = kwargs.get('req_keys', [])
    self.failfast = kwargs.get('failfast',False)
    self.data_key = kwargs.get('data_key')
    self.templatized = kwargs.get('templatized')
    self.config_file_auth_username = kwargs.get('auth_username')
    self.config_file_auth_password = kwargs.get('auth_password')
    self.configs_already_processed = []
    self.default_value = kwargs.get('default_value', {})




from btconfig.logger import Logger
from functools import reduce

# Setup Logging
logger = Logger().init_logger(__name__)

class AttrDict(dict):
    """
    see: https://stackoverflow.com/questions/3797957/python-easily-access-deeply-nested-dict-get-and-set
    """
    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise(TypeError, 'expected dict')

    def __setitem__(self, key, value):
        if '.' in key:
            myKey, restOfKey = key.split('.', 1)
            target = self.setdefault(myKey, AttrDict())
            if not isinstance(target, AttrDict):
                raise(KeyError, 'cannot set "%s" in "%s" (%s)' % (restOfKey, myKey, repr(target)))
            target[restOfKey] = value
        else:
            if isinstance(value, dict) and not isinstance(value, AttrDict):
                value = AttrDict(value)
            dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        if '.' not in key:
            return dict.__getitem__(self, key)
        myKey, restOfKey = key.split('.', 1)
        try:
            target = dict.__getitem__(self, myKey)
        except KeyError:
            return None
        return target[restOfKey]

    def __contains__(self, key):
        if '.' not in key:
            return dict.__contains__(self, key)
        myKey, restOfKey = key.split('.', 1)
        try:
            target = dict.__getitem__(self, myKey)
        except KeyError:
            return False
        return restOfKey in target

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default
        return self[key]

    def get(self, k, default=None):
        if '.' not in k:
            return super(AttrDict, self).get(k, default)
        myKey, restOfKey = k.split('.', 1)
        target = super(AttrDict, self).get(myKey, default)
        if not isinstance(target, AttrDict):
            return default
        return target.get(restOfKey, default)

    @staticmethod
    def merge(a, b, path=None):
        """merges b into a"""

        if not all([a,b]):
            return AttrDict(a)
        if path is None: path = []
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    AttrDict.merge(a[key], b[key], path + [str(key)])
                elif a[key] == b[key]:
                    pass  # same leaf value
                else:
                    pass
            else:
                a[key] = b[key]
        return AttrDict(a)

    def update(self, dict_path, default=None):
        """Interpret wildcard paths for setting values in a dictionary object"""
        result = {}
        if isinstance(self.dict, dict):
            result = reduce(lambda d, key: d.get(key, default) if isinstance(
                d, dict) else default, dict_path.split('.'), self.dict)

        if isinstance(result, dict):
            return AttrDict(result)
        else:
            return result

    __setattr__ = __setitem__
    __getattr__ = __getitem__



       



import logging
import logging.handlers
import os
import sys

class Logger:

  def __init__(self, **kwargs):
    self.debug = kwargs.get('debug', False)
    self.is_main_logger = kwargs.get('is_main_logger', False)

  def init_logger(self, name=None, debug=False):
    # Setup Logging
    logger = logging.getLogger(name)
    # TODO Find a better approach to this hacky method
    BTCONFIG_DEBUG = os.environ.get('BTCONFIG_DEBUG','').lower()
    if BTCONFIG_DEBUG in ['true','yes','on','1'] or self.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    streamhandler = logging.StreamHandler()
    streamhandler.setFormatter(
        logging.Formatter("%(asctime)s %(name)s [%(levelname)s]: %(message)s", datefmt='%Y-%m-%d %H:%M:%S')
    )
    # logger.propagate = False
    if self.is_main_logger:
      logger.addHandler(streamhandler)
    return logger


from btconfig import SuperDuperConfig
# Initialize Config Module
superconf = SuperDuperConfig()
# Initialize App Config
config = superconf.load_config('~/myconfig.yaml')
settings = superconf.get(config, 'section1.subsection1.item2')
print(settings)